/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 *
 *	Sound Emulation Routines,
 *	Martin GRIFFiths, 1995.
 *
 *	Last Updated : November 15th 1995.
 *
 * with modification by Dirk Jansen
 */

#include "main.h"
#include "tosdefs.h"
#include "mem.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <assert.h>
#include <sys/types.h>
#include "io.h"

#include "sb.h"

extern int snd_porta;

#include <sys/time.h>
#include <sys/ioctl.h>

#define DEBUG_SND 0

#define STATIC static

typedef struct
{	void (*machine_audio_open)(void);
	void (*machine_audio_close)(void);
	void (*machine_audio_volume)(void);
} Machine_Audio;


/*
 *	YM Audio Stuff.
 */

#define MAX_SAMPLES (8192)

#define ENV_SIZE (256)
#define NOISE_BITS (14)
#define NOISE_SIZE (1<<NOISE_BITS)

typedef struct 
{	unsigned int finetune;
	unsigned int crsetune;
	unsigned int amplitude;
} YM_Voice;

unsigned int YM_select;

struct Snd
{		
	YM_Voice YM_ChannelA; 
	YM_Voice YM_ChannelB; 
	YM_Voice YM_ChannelC; 
	unsigned int YM_NoiseControl;
	unsigned int YM_MixerControl;
	unsigned int YM_Envelope_fine_period;
	unsigned int YM_Envelope_crse_period;
	unsigned int YM_Envelope_shape;
	unsigned int IO_PortA;
	unsigned int IO_PortB;
	unsigned char YM_flag;
	unsigned char Server_Exit; 
} Snd;


/*
 * Hardware Register emulation functions 
 */ 
  
B LOAD_B_ff8800(void) 
{ 	
	/*fprintf(stderr,"Load $ff8800 - current selected : %d\n",YM_select);*/
	switch (YM_select)
	{	case 0:  
			return Snd.YM_ChannelA.finetune;
		case 1:
			return Snd.YM_ChannelA.crsetune;
		case 2:  
			return Snd.YM_ChannelB.finetune;
		case 3:
			return Snd.YM_ChannelB.crsetune;
		case 4:  
			return Snd.YM_ChannelC.finetune;
		case 5:
			return Snd.YM_ChannelC.crsetune;
		case 6:
			return Snd.YM_NoiseControl;
		case 7:
			return Snd.YM_MixerControl;
		case 8:
			return Snd.YM_ChannelA.amplitude;
		case 9:
			return Snd.YM_ChannelB.amplitude;
		case 10:
			return Snd.YM_ChannelC.amplitude;
		case 11:
			return Snd.YM_Envelope_fine_period;
		case 12:
			return Snd.YM_Envelope_crse_period;
		case 13:
			return Snd.YM_Envelope_shape;
		case 14:
			return snd_porta & 0xf8;
		case 15:
			return Snd.IO_PortB;
	}
	assert(0);
	return 0;
}

B LOAD_B_ff8801(void)
{	return (B) 0xff;			/* checked : always returns 0xff */
}

B LOAD_B_ff8802(void)
{	return (B) 0;
}

B LOAD_B_ff8803(void)
{	return (B) 0xff;			/* checked : always returns 0xff */
}

void STORE_B_ff8800(B v)
{	
	/*fprintf(stderr,"Store $ff8800 : %d\n",v);*/
	YM_select = v & 15;
}

void STORE_B_ff8801(B v)
{	
	/*fprintf(stderr,"Store $ff8801 : %d\n",v);*/

	/* should this do what 8800 does??? on a Falcon it seems to!!! */
}

void STORE_B_ff8802(B v)
{	
	/*fprintf(stderr,"Store $ff8802 : %d\n",v);*/
	switch (YM_select)
	{	case 0:  
			Snd.YM_ChannelA.finetune=v & 0xff;
#if (DEBUG_SND)
			fprintf(stderr,"Channel A fine tune:%d\n",(UL) v);
#endif
			break;	
		case 1:
			Snd.YM_ChannelA.crsetune=v & 0x0f; 
#if (DEBUG_SND)
			fprintf(stderr,"Channel A course tune:%d\n",(UL)v);
#endif
			break;					
		case 2:  
			Snd.YM_ChannelB.finetune=v & 0xff;
			break;	
		case 3:
			Snd.YM_ChannelB.crsetune=v & 0x0f;
			break;	
		case 4:  
			Snd.YM_ChannelC.finetune=v & 0xff;
			break;	
		case 5:
			Snd.YM_ChannelC.crsetune=v & 0x0f;
			break;	
		case 6:
			Snd.YM_NoiseControl=v & 0x1f;
			break;
		case 7:
			Snd.YM_MixerControl=v & 0xff;
#if (DEBUG_SND)
			fprintf(stderr,"Mixer Control:%d\n",(UL) v);
#endif
			break;
		case 8:
			Snd.YM_ChannelA.amplitude=v & 0x1f;
#if (DEBUG_SND)
			if (Snd.YM_ChannelA.amplitude & 0x10)
				fprintf(stderr,"Channel A Enveloping ON:%d\n",v);
#endif
			break;
		case 9:
			Snd.YM_ChannelB.amplitude=v & 0x1f;
#if (DEBUG_SND)
			if (Snd.YM_ChannelB.amplitude & 0x10)
				fprintf(stderr,"Channel B Enveloping ON:%d\n",v);
#endif
			break;
		case 10:
			Snd.YM_ChannelC.amplitude=v & 0x1f;
#if (DEBUG_SND)
			if (Snd.YM_ChannelC.amplitude & 0x10)
				fprintf(stderr,"Channel C Enveloping ON:\n");
#endif
			break;
		case 11:
			Snd.YM_Envelope_fine_period=v & 0xff;
#if (DEBUG_SND)
			fprintf(stderr,"Fine envelope:%d\n",(UL) v);
#endif
			break;
		case 12:
			Snd.YM_Envelope_crse_period=v & 0xff;
#if (DEBUG_SND)
			fprintf(stderr,"Course envelope:%d\n",(UL) v);
#endif
			break;
		case 13:
			Snd.YM_Envelope_shape=v & 0x0f;
#if (DEBUG_SND)
			fprintf(stderr,"Envelope shape set/trigger :%d\n",v);
#endif
			Snd.YM_flag = 1;
			break;
		case 14:
			snd_porta = v;
			break;
		case 15:
			Snd.IO_PortB = v;
			write_parallel(v);
			break;
	}
}

void STORE_B_ff8803(B v)
{	/*fprintf(stderr,"Store $ff8803 : %d\n",v);*/

	
}

static void init_noise(void);
static void init_envelope(void);
static void init_cliptable(void);

static void write(int dummy, void * buf, int bytes );

STATIC int devAudio;
unsigned char *audio_buffer;
STATIC int audio_buffsize; 
STATIC int audio_frequency;
STATIC int audio_samplebits;
int samples_to_generate;
STATIC int FREQ_base; 
extern long tv_usec;
extern int hz200pt;  
extern int verbose;


int STE_Microwire_volume = 64;
unsigned int YM_voice1_IntFrac;
unsigned int YM_voice2_IntFrac;
unsigned int YM_voice3_IntFrac;
unsigned int Env_IntFrac=0;
unsigned int Noise_IntFrac=0;

/* Envelope Definition stuff. */

typedef struct 
{	unsigned int Def[8];	
} Envelope_Definition;

STATIC Envelope_Definition Envelope00xx = {{ 1,0,0,0,0,0,0,0 }};
STATIC Envelope_Definition Envelope01xx = {{ 0,1,0,0,0,0,0,0 }};
STATIC Envelope_Definition Envelope1000 = {{ 1,0,1,0,1,0,1,0 }};
STATIC Envelope_Definition Envelope1001 = {{ 1,0,0,0,0,0,0,0 }};
STATIC Envelope_Definition Envelope1010 = {{ 1,0,0,1,1,0,0,1 }};
STATIC Envelope_Definition Envelope1011 = {{ 1,0,1,1,1,1,1,1 }};
STATIC Envelope_Definition Envelope1100 = {{ 0,1,0,1,0,1,0,1 }};
STATIC Envelope_Definition Envelope1101 = {{ 0,1,1,1,1,1,1,1 }};
STATIC Envelope_Definition Envelope1110 = {{ 0,1,1,0,0,1,1,0 }};
STATIC Envelope_Definition Envelope1111 = {{ 0,1,0,0,0,0,0,0 }};

STATIC Envelope_Definition *Envelopes[] = 
{
	&Envelope00xx,
	&Envelope00xx,
	&Envelope00xx,
	&Envelope00xx,
	&Envelope01xx,
	&Envelope01xx,
	&Envelope01xx,
	&Envelope01xx,
	&Envelope1000,
	&Envelope1001,
	&Envelope1010,
	&Envelope1011,
	&Envelope1100,
	&Envelope1101,
	&Envelope1110,
	&Envelope1111
};

STATIC int SquarewaveW[16] =
{ 	-128,-128,-128,-128,-128,-128,-128,-128,	
	+127,+127,+127,+127,+127,+127,+127,+127
}; 
STATIC int *clip_tab;
STATIC int clip_table[512];
STATIC unsigned char envs[16][ENV_SIZE*4];
STATIC char Noise_Table[NOISE_SIZE];
STATIC int Envelope[MAX_SAMPLES];
STATIC int Noise[MAX_SAMPLES];
STATIC int mix_buffer[MAX_SAMPLES];


static int amp_tab[16]= { 	 0, 1, 1, 2, 3,  4,  5,  7,	
				12,20,28,44,70,110,165,255	};

/* 
 *	Generate a single YM 2149 Voice.
 */

static void generate_YM_voice(YM_Voice *YM,unsigned int *Current_IntFrac,int TONE_MASK)
{ 	
	int i;
	unsigned int F=*Current_IntFrac;
	unsigned int YM_freq;
	unsigned int tune = YM->finetune + (YM->crsetune<<8);
	int *V = mix_buffer;
	if (tune)
	{	float t  = ((float)FREQ_base*65536*256*16) / tune;
		YM_freq = (unsigned int) t;
	}
	else
		YM_freq = 0;

	if (YM->amplitude & 0x10)	/* voice using Envelope? */
	{	
		if (Snd.YM_MixerControl & TONE_MASK)
		{	
			if (!(Snd.YM_MixerControl & (TONE_MASK<<3)))
			{	/* Has Envelope,No Tone, Has Noise */ 
				for (i = 0 ; i < samples_to_generate  ; i++)
					V[i] = clip_tab[V[i]+((Noise[i]*Envelope[i])>>8)];
			}
		} else
		{	
			if (Snd.YM_MixerControl & (TONE_MASK<<3))
			{ /* Has Envelope,Has Tone, No Noise */
				for (i = 0 ; i < samples_to_generate ; i++,F+=YM_freq)
					V[i] = clip_tab[V[i]+((SquarewaveW[F>>28]*Envelope[i])>>8)];
			} else
			{ /* Has Envelope,Has Tone and Noise */ 
				for (i = 0 ; i < samples_to_generate  ; i++,F+=YM_freq)
					V[i] = clip_tab[V[i]+((clip_tab[SquarewaveW[F>>28]+Noise[i]]*Envelope[i])>>8)];
 			}
		}
	} else				/* voice using amplitude */
	{	
#if (1)
		int amp = amp_tab[YM->amplitude];
#else
		int amp = (YM->amplitude) * 16;
#endif
		if (Snd.YM_MixerControl & TONE_MASK)
		{	
			if (!(Snd.YM_MixerControl & (TONE_MASK<<3)))
			{	/* No Envelope, No Tone, Has Noise */ 
				for (i = 0 ; i < samples_to_generate  ; i++)
					V[i] = clip_tab[V[i]+((Noise[i]*amp)>>8)];
			}	
		} else
		{	
			if (Snd.YM_MixerControl & (TONE_MASK<<3))
			{ /* No Envelope, Has Tone, No Noise */
				for (i = 0 ; i < samples_to_generate ; i++,F+=YM_freq)
					V[i] = clip_tab[V[i]+((SquarewaveW[F>>28]*amp)>>8)]; 
			} else
			{ /* No Envelope, Has Tone, Has Noise */ 
				for (i = 0 ; i < samples_to_generate  ; i++,F+=YM_freq)
					V[i] = clip_tab[V[i]+((clip_tab[SquarewaveW[F>>28]+Noise[i]]*amp)>>8)];
			} 
		} 
	}	 
	*Current_IntFrac = F; 
} 

static int end_generate_YM_voice;

/* 
 *	The Audio Server... 
 */ 

static int audio_server(void)
{	register int i;

		if (Snd.YM_flag) Env_IntFrac=0; 

		if ((Snd.YM_MixerControl & 0x38) != 0x38) /*any noise channels? */ 
		{ 	unsigned int Noise_Freq;
			if (Snd.YM_NoiseControl > 0)
			{	Noise_Freq = (FREQ_base<<(4+8))/(Snd.YM_NoiseControl);
			} else
				Noise_Freq = 0;
			for (i = 0 ; i < samples_to_generate ; i++,Noise_IntFrac+=Noise_Freq)
				Noise[i] = (int) Noise_Table[(Noise_IntFrac >> 16) & (NOISE_SIZE-1)];
		}

		if ( (Snd.YM_ChannelA.amplitude & 0x10) || (Snd.YM_ChannelB.amplitude & 0x10) || (Snd.YM_ChannelC.amplitude & 0x10) ) 
		{	/* Generate Frequency Shifted Envelope */
			unsigned char *Env= &envs[Snd.YM_Envelope_shape][0];
			unsigned int env_period = (Snd.YM_Envelope_crse_period<<8) + Snd.YM_Envelope_fine_period;
			unsigned int Env_Freq;
			if (env_period > 0)
			{	Env_Freq = (FREQ_base<<16) / env_period;
			} else
				Env_Freq = 0;

			for (i = 0 ; i < samples_to_generate ; i++)
			{	Envelope[i] =  amp_tab[Env[Env_IntFrac>>16]>>4];
				Env_IntFrac += Env_Freq;
				/* this is particularly crap... */
				while (Env_IntFrac >= ((ENV_SIZE*4)<<16)) 
					Env_IntFrac -= ((ENV_SIZE*2)<<16);	
			}
		}

		bzero(mix_buffer,samples_to_generate*sizeof(int)); 

		generate_YM_voice(&Snd.YM_ChannelA,&YM_voice1_IntFrac,1); 
		generate_YM_voice(&Snd.YM_ChannelB,&YM_voice2_IntFrac,2);
		generate_YM_voice(&Snd.YM_ChannelC,&YM_voice3_IntFrac,4);
		if (audio_samplebits == 8)
		{	char *buf = (char *) audio_buffer;
   		for (i = 0 ; i < samples_to_generate; i++) {
				buf[i] = mix_buffer[i]^0x80;
         }
		} else
		{	
			UW *buf = (UW *) audio_buffer;
			for (i = 0 ; i < samples_to_generate ; i++) 
				buf[i] = (mix_buffer[i]<<6)^0x8000; 
		}
      return 0;
}

static int end_audio_server;
/*
 *	Init clip table 
 */

void init_cliptable(void)
{	int i;	
	for (i=0 ; i < 512 ; i++)
	{	int j = i - 256;	
		clip_table[i]= j<-128 ? -128 : j>127 ? 127 : j; 
	}
	clip_tab = &clip_table[256];
}

/*
 *	Create a 'noise' sample.
 */

static void init_noise(void)
{	int i;
	srand(0xa354e767);
	for (i = 0 ; i < NOISE_SIZE ; i++)
		Noise_Table[i] = (rand() & 0xff)-128;
}

/*
 *	Initialise tables for the '16' YM Amplitude Envelopes.
 */
 
static void init_envelope(void)
{	int i,j,k;
	for (i = 0 ; i < 15 ; i++)
	{	for (j = 0 ; j < 4; j++)
		{ 	int Env_Src = Envelopes[i]->Def[2*j]*(ENV_SIZE-1);
			int Env_Dst = Envelopes[i]->Def[2*j+1]*(ENV_SIZE-1);
			int Env_Cur = Env_Src;
			for (k = 0 ; k < ENV_SIZE ; k++)
			{	envs[i][(j*ENV_SIZE)+k] = Env_Cur;
				if (Env_Dst > Env_Src)
					Env_Cur++;
				else if (Env_Dst < Env_Src)
					Env_Cur--;
			}
		}
	}
}

/* 
 *	DOS Specific...
 */

static int audio_ok = 0;

void audio_irq ( void )
{

   audio_server();
	Snd.YM_flag = 0;

   if ( playingpart == 0 )
       dosmemput ( audio_buffer, samplesize, dmabufaddr+samplesize );
   else
       dosmemput ( audio_buffer, samplesize, dmabufaddr );

}

static int end_audio_irq;

int lock_memory(void *ptr, int size, int code);

void audio_open(void)
{
 	init_noise();
 	init_envelope();
 	init_cliptable();

   audio_ok = 0;

   lock_memory ( audio_server, (int)&end_audio_server-(int)audio_server, 1 );
   lock_memory ( generate_YM_voice, (int)&end_generate_YM_voice-(int)generate_YM_voice, 1 );
   lock_memory ( audio_irq, (int)&end_audio_irq-(int)audio_irq, 1 );
   lock_memory ( &Snd, sizeof(Snd), 0 );
   lock_memory ( &Env_IntFrac, sizeof(Env_IntFrac), 0 );
   lock_memory ( &samples_to_generate, sizeof(samples_to_generate), 0 );
   lock_memory ( &Noise_IntFrac, sizeof(Noise_IntFrac), 0 );
   lock_memory ( Noise_Table, sizeof(Noise_Table), 0 );
   lock_memory ( Noise, sizeof(Noise), 0 );
   lock_memory ( &audio_buffer, sizeof(audio_buffer), 0 );
   lock_memory ( &audio_samplebits, sizeof(audio_samplebits), 0 );
   lock_memory ( &YM_voice1_IntFrac, sizeof(YM_voice1_IntFrac), 0 );
   lock_memory ( &YM_voice2_IntFrac, sizeof(YM_voice2_IntFrac), 0 );
   lock_memory ( &YM_voice3_IntFrac, sizeof(YM_voice3_IntFrac), 0 );
   lock_memory ( mix_buffer, sizeof(mix_buffer), 0 );
   lock_memory ( amp_tab, sizeof(amp_tab), 0 );
   lock_memory ( Envelope, sizeof(Envelope), 0 );
   lock_memory ( envs, sizeof(envs), 0 );
   lock_memory ( &FREQ_base, sizeof(FREQ_base), 0 );
   lock_memory ( clip_table, sizeof(clip_table), 0 );
   lock_memory ( &clip_tab, sizeof(clip_tab), 0 );
   lock_memory ( SquarewaveW, sizeof(SquarewaveW), 0 );
   lock_memory ( &playingpart, sizeof(playingpart), 0 );
   lock_memory ( &samplesize, sizeof(samplesize), 0 );
   lock_memory ( &dmabufaddr, sizeof(dmabufaddr), 0 );
//   lock_memory ( dosmemput, 10000, 1 );
//   lock_memory ( movedata, 10000, 1 );
//   lock_memory ( _dj_movedata, 10000, 1 );

	if ((audio_buffer =
		(unsigned char *)malloc(audio_buffsize * sizeof(char))) == NULL)
	{
		fprintf(stderr,"DOS audio: Failed to allocate audio buffer memory\n");
      exit_sb ();
		exit(1);
	}

   lock_memory ( audio_buffer, audio_buffsize, 0 );

   audio_samplebits = 8;

	/* Get the Size of Audio Hardware Buffer */
	
	audio_buffsize = samplesize;

	if (verbose)
		fprintf(stderr,"DOS audio: hardware buffer size:%d\n",audio_buffsize);

	/* Setup Hardware replay frequency */

	audio_frequency = samplefreq;
	if (verbose)
		fprintf(stderr,"DOS audio: sound frequency:%d\n",audio_frequency);

   FREQ_base = 60;

	samples_to_generate = samplesize;

   if ( !init_sb () ) {
       fprintf ( stderr, "DOS audio: cannot init audio...\n" );
       fprintf ( stderr, "           %s\n", last_error_msg );
       return;
   }/*if*/

   fprintf ( stderr,"DOS audio: using SB at port %X, irq %d and dma %d.\n", sb_baseaddr, sb_irq, sb_dmachannel );

	if (verbose)
		fprintf(stderr,"DOS audio: sound configuration successful.\n");

    audio_ok = 1;
}

void audio_close(void)
{
    audio_ok = 0;

    exit_sb ();
    fprintf ( stderr, "DOS audio: audio removed.\n" );
}


