#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <go32.h>
#include <dpmi.h>
#include <dos.h>
#include <conio.h>
#include <bios.h>
#include <stdarg.h>

#include "main.h"
#include "dosscan.h"

#define DOS_DEBUG_FILE "debug.txt"

#define DEBUG_COL 15+16
#define INFO_COL 14+16*(2+4)

static char buffer[2048];
static char screen[80*50];
static char info[80*50];

extern char volatile keyflag[128];
extern int hz200pt, vblpt;

extern int refresh;
extern int shiftmod;

static void mem2scrcpy (void * dest, void * sour, int i)
{
    movedata ( _go32_my_ds(), (unsigned) (sour), _go32_conventional_mem_selector(), (unsigned) dest, i );
}

static void textout ( int x, int y, char * str ) {

    char *p = info+y*160+x*2;
    int i = 0;

    while ( str[i] != 0 )
    {
        *(p++) = str[i++];
        *(p++) = INFO_COL;
    }
}

void init_debug_screen ( void ) {
    int i;
    FILE *f;

    f = fopen ( DOS_DEBUG_FILE, "w" );
    if ( f ) fclose ( f );

    for ( i=0; i<80*25*2; i+=2 ) {
        screen[i] = 0; screen[i+1] = DEBUG_COL;
        info[i] = 0; info[i+1] = INFO_COL;
    }/*for*/

    i = 1;
    textout ( 1,i++, "press [M] - mono mode" );
    textout ( 1,i++, "      [C] - color mode" );
    i++;
    textout ( 1,i++, "      [1] - decrease VBL timing" );
    textout ( 1,i++, "      [2] - increase VBL timing" );
    i++;
    textout ( 1,i++, "      [3] - decrease screen refresh rate" );
    textout ( 1,i++, "      [4] - increase screen refresh rate" );
    i++;
    textout ( 1,i++, "      [S] - make screen shot" );
    textout ( 1, 23, " press [ESC] to switch back to emulation" );
}

void show_debug_screen ( void ) {
    union REGS r;
    int old_video_modus;
    int fertig = 0;

    // save old video mode
    r.h.ah = 15;
    int386 ( 0x10, &r, &r );
    old_video_modus = r.h.al;

    r.w.ax = 0x1202;
    r.w.bx = 0x30;
    int386(0x10,&r,&r);
    r.w.ax = 0x3;
    int386(0x10,&r,&r);
    r.w.ax = 0x1112;
    r.w.bx = 0x0;
    int386(0x10,&r,&r);


    mem2scrcpy ( (void*) 0xb8000, screen, 80*25*2 );
    mem2scrcpy ( (void*) 0xb8000+80*25*2, info, 80*25*2 );

    while ( !fertig )
    {
        // esc
        if ( keyflag[1] == 0 ) {
            while ( keyflag[1] == 0 ); fertig = 1;
        }

        // 0
        if ( keyflag[SCANCODE_C] == 0 )
        {
            while ( keyflag[SCANCODE_C] == 0 );
            gr_mode = MODE_COLOR;
            shiftmod = 0;
            fprintf ( stderr, "switched to color mode\n" );
            fertig = 1;
            init_mem();
            init_cpu();
        }
        if ( keyflag[SCANCODE_M] == 0 )
        {
            while ( keyflag[SCANCODE_M] == 0 );
            gr_mode = MODE_MONO;
            shiftmod = 2;
            fprintf ( stderr, "switched to mono mode\n" );
            fertig = 1;
            init_mem();
            init_cpu();
        }
        if ( keyflag[SCANCODE_2] == 0 )
        {
            while ( keyflag[SCANCODE_2] == 0 );
            if ( vblpt > 1 ) {
                vblpt--;
                print_timing();
                mem2scrcpy ( (void*) 0xb8000, screen, 80*25*2 );
            }
        }
        if ( keyflag[SCANCODE_1] == 0 )
        {
            while ( keyflag[SCANCODE_1] == 0 );
            if ( vblpt < 10 ) {
                vblpt++;
                print_timing();
                mem2scrcpy ( (void *) 0xb8000, screen, 80*25*2 );
            }
        }
        if ( keyflag[SCANCODE_4] == 0 )
        {
            while ( keyflag[SCANCODE_4] == 0 );
            if ( refresh > 1 ) {
                refresh--;
                print_timing();
                mem2scrcpy ( (void *) 0xb8000, screen, 80*25*2 );
            }
        }
        if ( keyflag[SCANCODE_3] == 0 )
        {
            while ( keyflag[SCANCODE_3] == 0 );
            if ( refresh < 10 ) {
                refresh++;
                print_timing();
                mem2scrcpy ( (void *) 0xb8000, screen, 80*25*2 );
            }
        }

        if ( keyflag[SCANCODE_S] == 0 )
        {
            while ( keyflag[SCANCODE_S] == 0 );
            save_pcx();
            mem2scrcpy ( (void *) 0xb8000, screen, 80*25*2 );
       }

    }

    r.w.ax = 0x1202;
    r.w.bx = 0x30;
    int386(0x10,&r,&r);

    // restore old video modus
    r.w.ax = old_video_modus;
    int386 ( 0x10, &r, &r );

}

void print ( char * text ) {
    char *t = text;
    static int x = 0, y = 0;
    int i;

    while ( *t != 0 )
    {
        if ( *t == '\n' )
        {
            x = 0;
            y++;
        }/*if*/
        else {
            *(screen+y*80*2+x*2) = *t;
            *(screen+y*80*2+x*2+1) = DEBUG_COL;
            x++;
            if ( x > 79 )
            {
                x = 0;
                y++;
            }/*if*/
        }/*else*/
        while ( y > 24 )
        {
            y--;
            memcpy ( screen, screen+160, 160*24 );
            for ( i=0; i<=79; i++) *(screen+160*24+i*2) = 0;
        }/*if*/
        t++;
    }/*while*/
}

static void add_debug_file ( char * text ) {
    FILE * f;

    f = fopen ( DOS_DEBUG_FILE, "a" );
    if ( f )
    {
        fwrite ( text, 1, strlen ( text ), f );
        fflush ( f );
        fclose ( f );
    }
}

int fprintf( FILE *fp, const char *f, ...)
{
    va_list argptr;                /* -> variable argument list */
    long count;                   /* count of characters formatted */

    va_start(argptr, f);           /* get pointer to argument list */
    count=vsprintf(buffer, f, argptr);
    va_end(argptr);                /* done with variable arguments */

    if ( fp == stderr )
    {
        print ( buffer );
        add_debug_file ( buffer );

        if (count<0) return count-1000;/* bad start */
        buffer[count]='\0';            /* ensure terminated */
        return count;                  /* all formatted data processed */
    }
    else return fwrite ( buffer, 1, strlen ( buffer ),fp );

} /* printf */