#include <stdio.h>
#include <errno.h>
#include <unistd.h>

#include "missing.h"
#include "defs.h"
#include "cpu.h"
#include "main.h"
#include "mem.h"
#include "screen.h"

extern int scr_width,scr_height;

#define IntSwap(x) (((x&0xff)<<8)+((x>>8)&0xff))

/* save_tga:
 *  Writes a bitmap into a TGA file, using the specified pallete (this 
 *  should be an array of at least 256 RGB structures).
 */
int save_pcx(void)
{
   unsigned char image_palette[256][3];
   unsigned int y,x,i=0,j=0,k=0;
   unsigned short int c,w1,w2,w3,w4;
   int w, h, nr = 0;
   char *p;
   unsigned short int *p2;
   FILE *f;
   static char colconv[]={0, 36, 73, 109, 145, 181, 218, 255};
   char filename[20];
   int runcount;
   char runchar;

   if ( shiftmod == 2 ) {
      c = LM_UW(&color[0]) & 1;
      i = c ? 255 : 0;
      j = i ? 0 : 255;
      image_palette[0][2] = i;
      image_palette[0][1] = i;
      image_palette[0][0] = i;
      image_palette[1][2] = j;
      image_palette[1][1] = j;
      image_palette[1][0] = j;
   }
   else {
      for(y=0; y<256; y++) {
         c = LM_UW(&color[y&15]);
         image_palette[y][2] = colconv[((c>>8)&7)];;
         image_palette[y][1] = colconv[((c>>4)&7)];;
         image_palette[y][0] = colconv[(c&7)];;
      }
   }

   // create filename
   do {
       sprintf ( filename, "file%4.4x.pcx", nr++ );
   } while (access ( filename, F_OK ) == 0 && nr<60000 );

   f = fopen( filename, "wb" );
   if (!f) {
      fprintf ( stderr, "cannot create screen shot! (%d)\n", errno );
      return errno;
   }
   fputc(10, f);                /* manufacturer */
   fputc(5, f);                 /* version */
   fputc(1, f);                 /* run length encoding  */
   fputc(8, f);                 /* 8 bits per pixel */
   fputc(0, f);fputc(0,f);                /* xmin */
   fputc(0, f);fputc(0,f);                /* ymin */
   w = scr_width-1;
   h = scr_height-1;
   fputc(w&0xff, f);fputc(w>>8,f);     /* xmax */
   fputc(h&0xff, f);fputc(h>>8,f);     /* ymax */
   w = 320;
   h = 200;
   fputc(w&0xff, f);fputc(w>>8,f);     /* HDpi */
   fputc(h&0xff, f);fputc(h>>8,f);     /* VDpi */
   for (c=0; c<16; c++) {
      fputc(image_palette[c][0], f);
      fputc(image_palette[c][1], f);
      fputc(image_palette[c][2], f);
   }
   fputc(0, f);                 /* reserved */
   fputc(1, f);                 /* one color plane */
   w = scr_width;
   fputc(w&0xff, f);fputc(w>>8,f);     /* number of bytes per scanline */
   h = 1;
   fputc(h&0xff, f);fputc(h>>8,f);     /* color palette */
   w = scr_width;
   h = scr_height;
   fputc(w&0xff, f);fputc(w>>8,f);     /* hscreen size */
   fputc(h&0xff, f);fputc(h>>8,f);     /* vscreen size */
   for (c=0; c<54; c++)             /* filler */
      fputc(0, f);

   p = MEM(vbase);
   for(y=0; y<h; y++) {
      runcount = 0;
      runchar = 0;
      for(x=0; x<w; x++) {
          switch ( shiftmod ) {
              case 0:
                      i=x/16;
                      i=i*4;
                      p2 = (unsigned short int *) p+y*w/4;
                      w1 = IntSwap(p2[i]);
                      w2 = IntSwap(p2[i+1]);
                      w3 = IntSwap(p2[i+2]);
                      w4 = IntSwap(p2[i+3]);
                      j = 15-(x%16);
                      w1 = ( w1 >> j ) & 1;
                      w2 = ( w2 >> j ) & 1;
                      w3 = ( w3 >> j ) & 1;
                      w4 = ( w4 >> j ) & 1;
                      c = (w4 << 3) + ( w3 << 2 ) + ( w2 << 1 ) + w1;
                      break;
              case 1:
                      i=x/16;
                      i=i*2;
                      p2 = (unsigned short int *) p+y*w/8;
                      w1 = IntSwap(p2[i]);
                      w2 = IntSwap(p2[i+1]);
                      j = 15-(x%16);
                      w1 = ( w1 >> j ) & 1;
                      w2 = ( w2 >> j ) & 1;
                      c = ( w2 << 1 ) + w1;
                      break;
              case 2: i = x/8;
                      j = x%8;
                      if ( p[y*w/8+i] & (1<<(7-j))) c = 1;
                      else c = 0;
                      break;
              default:
                      c = 0;
                      break;
          }/*switch*/
          if (runcount==0) {
	          runcount = 1;
	          runchar = c;
       	 }/*if*/
          else {
	            if ((c != runchar) || (runcount >= 0x3f)) {
	               if ((runcount > 1) || ((runchar & 0xC0) == 0xC0))
		               fputc(0xC0 | runcount, f);
                  fputc(runchar,f);
	               runcount = 1;
	               runchar = c;
               }/*if*/
	            else
	                runcount++;
          }/*else*/
      }/*for*/
      if ((runcount > 1) || ((runchar & 0xC0) == 0xC0))
	      fputc(0xC0 | runcount, f);
      fputc(runchar,f);
   }/*for*/
   fputc(12,f);                 /* 256 color pallete flag */
   fwrite(image_palette, 1, 768, f);
   fclose(f);
   fprintf ( stderr, "screen shot written to %s.\n", filename );
   return errno;
}




