#include <stdio.h>
#include <string.h>
#include <go32.h>
#include <dpmi.h>
#include <sys/nearptr.h>
#include "missing.h"
#include "defs.h"
#include "cpu.h"
#include "main.h"
#include "mem.h"
#include "screen.h"

#include <dos.h>
#include <bios.h>
#include <conio.h>
// #include <i86.h>
#include <stdlib.h>
#include <setjmp.h>

#include "ikbd.h"
#include "dosscan.h"
#include "sb.h"
#include "dosdebug.h"

int scr_width,scr_height,scr_planes;

static UW old_color[16];        /* last known color palette */

int do_exit = 0;
int do_debug = 0;

// **************
// keyboard stuff
// **************

static int dos_keytab[][2] =
{
	{ SCANCODE_INS,ST_INSERT },
	{ SCANCODE_DEL,ST_DELETE },
    { SCANCODE_ISO, ST_ISO },
    { SCANCODE_KEYPADENTER,ST_KP_ENTER },
    { SCANCODE_KEYPADMULT,ST_KP_MULT },
    { 70,ST_HELP }, // Rollen
	{ SCANCODE_PAUSE,ST_UNDO },
	{ SCANCODE_PGUP,ST_KP_OPEN },
	{ SCANCODE_PGDOWN,ST_KP_CLOSE },
    { -1, -1 }
};

static char conv_tab[128];

static void keyboard_handler ( void );

static _go32_dpmi_seginfo kb_old;   /* original prot-mode key IRQ */
static _go32_dpmi_seginfo kb_new;   /* prot-mode interrupt segment info*/

char volatile keyflag[128];
int volatile lastcode = 0;
static int last_ext=0;

#define EXIT_KEYBHANDLER enable();outportb(0x20,0x20);return;

static void keyboard_handler ( void )
{
    int temp, scancode, state;

    temp = inp(0x60);
    disable();

    // erweiterte Tasten kommt rein
    if ( last_ext == 1 )
    {
        last_ext = 0;
    }

    // gleich kommt eine erweiterte Taste
    if ( temp == 224 ) {
        last_ext = 1;
        EXIT_KEYBHANDLER;
    }
    else last_ext = 0;

    scancode = temp & 0x7f;
//    printf ( "scancode: %d  \r", scancode );
//    fflush ( stdout );
    state = temp & 0x80;

    if ( keyflag[scancode] == state ) {
        EXIT_KEYBHANDLER;
    }/*if*/
    else keyflag[scancode] = state;

    // if we are in debug screen, do not send keys to ST
    if ( do_debug ) { EXIT_KEYBHANDLER; }

    switch ( scancode )
    {
        // F11
        case 87:
            if ( state == 0 )
            {
                do_debug = 1;
                EXIT_KEYBHANDLER;
            }/*if*/
            else EXIT_KEYBHANDLER;
        // F12
        case 88:
            do_exit = 1;
            EXIT_KEYBHANDLER;
    }
    lastcode = scancode;
    scancode = conv_tab[scancode];
#if 0
    printf ( "\r scancode: %d --> %d ", lastcode, scancode );
    fflush ( stdout );
#endif
    ikbd_send ( scancode | state );
    EXIT_KEYBHANDLER;
}

// mark for calculation of size of keyboard interrupt
static int kb_int_end;

int lock_memory(void *ptr, int size, int code)
{
    __dpmi_meminfo meminfo;
    int seg = (code ? _my_cs() : _my_ds());

    if (__dpmi_get_segment_base_address(seg, &meminfo.address) != 0){
        return -1;
    }

    meminfo.size = size;
    meminfo.address += (long)ptr;
    return __dpmi_lock_linear_region(&meminfo);
}

static void init_keyboard ( void ) {
    int i;

    // init keyflag array
    memset ( (char*) keyflag, 0x80, sizeof ( keyflag ) );
    // init conv_tab
    for ( i=0; i<sizeof(keycodes)/sizeof(keycodes[0]);i++) keycodes[i] = i;
    for ( i=0; i<sizeof(conv_tab)/sizeof(conv_tab[0]);i++) conv_tab[i] = i;
    for ( i=0; dos_keytab[i][0] != -1; i++ ) {
        if ( conv_tab[dos_keytab[i][0]] != dos_keytab[i][0] )
        {
            fprintf ( stderr, "warning! tried to assign a key twice (%d)!\n", dos_keytab[i][0] );
        }
        else {
            conv_tab[dos_keytab[i][0]] = dos_keytab[i][1];
            fprintf ( stderr, "assigned PC keycode %d to Atari keycode %d\n", dos_keytab[i][0], dos_keytab[i][1] );
        }
    }
    // install new keyboard handler

    lock_memory( (void *) keyflag, (int)sizeof(keyflag), 0);
    lock_memory( &do_debug, (int)sizeof(do_debug), 0);
    lock_memory( conv_tab, (int)sizeof(conv_tab), 0);
    lock_memory( &last_ext, (int)sizeof(last_ext), 0);
    lock_memory( (void *) &lastcode, (int)sizeof(lastcode), 0);
    lock_memory( keyboard_handler, (int)&kb_int_end-(int)keyboard_handler, 1);

    fflush(stdin);

    kb_new.pm_offset = (int)keyboard_handler;       /* pm handler */
    _go32_dpmi_allocate_iret_wrapper(&kb_new);
    kb_new.pm_selector = _go32_my_cs();
    _go32_dpmi_get_protected_mode_interrupt_vector(9, &kb_old);
    _go32_dpmi_set_protected_mode_interrupt_vector(9, &kb_new);

    fprintf ( stderr, "keyboard handler installed\n" );
}

static void exit_keyboard ( void ) {
    // restore old keyboard handler
    _go32_dpmi_set_protected_mode_interrupt_vector(9, &kb_old);
    _go32_dpmi_free_iret_wrapper(&kb_new);
}

// ***********
// mouse stuff
// ***********

int mouse_present = 0;

static void init_mouse ( void ) {
    union REGS r;
    static int first_time = 1;

    // look if mouse is present
    r.x.ax = 0;
    int386(0x33, &r, &r );
    if ( r.x.ax != 0xffff ) {
        if ( first_time ) fprintf ( stderr, "no mouse found. Install mouse driver!\n" );
    }
    else {
        mouse_present = 1;
        if ( first_time ) fprintf ( stderr, "mouse installed.\n" );

        // set mouse movement area
        r.x.ax = 7;
        r.x.cx = 0;
        r.x.dx = (640-1)*8;
        int386 (0x33, &r, &r);

        r.x.ax = 8;
        r.x.cx = 0;
        r.x.dx = (400-1)*8;
        int386 (0x33, &r, &r);

        r.x.ax = 2;
        int386 (0x33, &r, &r);

        r.x.ax = 15;
        r.x.cx=r.x.dx=2;
        int386 (0x33, &r, &r);

    }/*else*/
    first_time = 0;
}

static void exit_mouse ( void ) {
}

static void get_mouse_status ( void ) {
    union REGS r;
    int b;
    long xdiv, ydiv;
    static oldb = -1;

    // get Button state
    r.x.ax = 0x0003;
    int386 ( 0x33, &r, &r );
    b = r.w.bx;
    if ( oldb != b )
    {
        ikbd_button(1, (b & 1) != 0);
        ikbd_button(2, (b & 2) != 0);
        oldb = b;
    }/*if*/
    r.x.ax = 0x000b;
    int386 ( 0x33, &r, &r );
    xdiv = r.w.cx; if ( xdiv >= 32768 ) xdiv = -(65536-xdiv);
    ydiv = r.w.dx; if ( ydiv >= 32768 ) ydiv = -(65536-ydiv);
    if ( xdiv != 0 || ydiv != 0 ) ikbd_adjust ( xdiv/2, ydiv/2 );
}

// ************
// screen stuff
// ************

extern int custom_screensize;

inline void dos_memset ( void * dest, unsigned char i, int l )
{
    __djgpp_nearptr_enable();
    memset ( dest + __djgpp_conventional_base, i, l );
    __djgpp_nearptr_disable();
}

static int init800x600 ( void ) {
    union REGS r;
    int i = -1;
    static unsigned char modenr[]={0x6a, 0x58, 0x29, 0x54, 0x16, 0x79, 255, 255};

    do {
        i++;
        r.h.ah = 0;
        r.h.al = modenr[i];
        int386 ( 0x10, &r, &r );
        r.h.ah = 0x0f;
        int386 ( 0x10, &r, &r );
    } while ( r.h.al != modenr[i] && modenr[i+1] != 255 );
    if ( modenr[i] == r.h.al ) return 1;
    else return 0;
}

static void setpalettecolor ( int c, int r, int g, int b ) {
    if ( c == -1 ) return;
    outp ( 0x3c8, c);
    outp ( 0x3c9, r);
    outp ( 0x3c9, g);
    outp ( 0x3c9, b);
}

static void change_colors (void)
{   int i, j;
    static int lowpalconv[]= {0 , 1, 2, 3 , 4, 5, 6, 7, 16, 17, 18, 19, 20, 21, 22, 23};
    static int custpalconv[]={0 , 1, 2, 3 , 4, 5, 20, 7, 56, 57, 58, 59, 60, 61, 62, 63};
    static char colconv[]={0, 9, 18, 27, 36, 45, 54, 63 };
    static int old_mono_inv = 2;

    if (shiftmod == 2 )
    {
        i = old_color[0] & 1;
        i = i ? 63 : 0;
        j = i ? 0 : 63;
        if ( !custom_screensize )
        {
            setpalettecolor(0,i,i,i); setpalettecolor(63,j,j,j);
        }/*if*/
        else {
            setpalettecolor(0,i,i,i); setpalettecolor(custpalconv[1],j,j,j);
        }/*else*/
        if ( ( old_color[0] & 1 ) != old_mono_inv && !custom_screensize )
        {
            old_mono_inv = old_color[0] & 1;
            i = old_mono_inv ? 0xff : 0;
            dos_memset ( (void *) 0xa0000, i, 640*40/8 );
            dos_memset ( (void *) 0xa0000+640*440/8, i, 640*40/8 );
        }/*if*/
    }/*if*/
    else {
        for (i=0; i<COLS; i++)
        {
            int r,g,b,c;
            c = old_color[i];
            #if !STE
            c &= 0x777;
            #endif
            r       = colconv[((c>>8)&7)];  // <<3)|((c>>11)<<2);
            g       = colconv[((c>>4)&7)];  // <<3)|(((c>>7)&1)<<2);
            b       = colconv[(c&7)];       // <<3)|(((c>>3)&1)<<2);
            if ( !custom_screensize ) setpalettecolor (lowpalconv[i], r, g, b);
            else setpalettecolor (custpalconv[i], r, g, b);
        }
    }
}

static int old_video_modus;

void dos_screen_open ( void ) {
    union REGS r;

    // save old video mode
    r.h.ah = 15;
    int386 ( 0x10, &r, &r );
    old_video_modus = r.h.al;

    init_mouse ();
    init_keyboard ();

}

void dos_screen_close ( void ) {
    union REGS r;

    exit_keyboard ();
    exit_mouse ();

    // read actual video mode
    r.h.ah = 15;
    int386 ( 0x10, &r, &r );
    // restore only if video mode was changed
    if ( r.h.al != old_video_modus ) {
        // restore old video modus
        r.x.ax = old_video_modus;
        int386 ( 0x10, &r, &r );
    }
}

int refcount = 0;
extern jmp_buf env;

static inline void hires ( char * sour, char * dest, unsigned long int i)
{
    dosmemput(sour, i, (int) dest);
}

static inline void medres ( void * sour, void * dest, unsigned long int i)
{
    __djgpp_nearptr_enable();
    dest += __djgpp_conventional_base;
    while (i>0)
    {
        *((short int *)dest) = *((short int *)sour);
        sour+=4; dest+=2; i--;
    }
    __djgpp_nearptr_disable();
}

static inline void lowres ( char * sour, char * dest, unsigned long int i)
{
    __djgpp_nearptr_enable();
    dest += __djgpp_conventional_base;
    while (i>0)
    {
        *((short int *)dest) = *((short int *)sour);
        sour+=8; dest+=2; i--;
    }
    __djgpp_nearptr_disable();
}

static void change_mode ( void ) {
    union REGS r;

    scr_width = scr_def[shiftmod].w;
    scr_height = scr_def[shiftmod].h;
    scr_planes=4>>shiftmod;

    if ( !custom_screensize )
    {
        switch (shiftmod)
        {
            case 0:
                r.x.ax = 0xd;
                int386 ( 0x10, &r, &r );
                break;
            case 1:
                r.x.ax = 0xe;
                int386 ( 0x10, &r, &r );
                break;
            case 2:
                r.x.ax = 0x11;
                int386 ( 0x10, &r, &r );
                // fill background
                dos_memset ( (void *) 0xa0000, 0xff, 640*480/8 );
                // set screen
                hires ( MEM(vbase), (char*) 0xa0000+40*640/8, 640*400/8 );
                break;
        }/*switch*/
    }
    else
    {
         init800x600 ();
         outp ( 0x3c4, 2); outp ( 0x3c5, 1+2+4+8 );  // alle Planes selektieren
         dos_memset ( (void *) 0xa0000, 0, 800*600/8 );
    }

    change_colors();
    init_mouse ();
}

void dos_screen_shifter ( void ) {
    int w, pal = 0, i;
    unsigned char *p;
    static old_shiftmod = -1;

    // jump back to main if exit requested
    if ( do_exit ) longjmp ( env, 1 );
    if ( do_debug ) {
        set_sb_speaker ( 0 );
        show_debug_screen();
        set_sb_speaker ( 1 );
        do_debug = 0;
        change_colors();
        init_mouse ();
        if ( shiftmod == 2 && !custom_screensize ) {
            // fill background
            i = ( old_color[0] & 1 ) ? 0xff : 0;
            dos_memset ( (void *) 0xa0000, i, 640*40/8 );
            dos_memset ( (void *) 0xa0000+640*440/8, i, 640*40/8 );
        }/*if*/
        if ( shiftmod == 2 && custom_screensize ) {
            // fill background
            dos_memset ( (void *) 0xa0000, 0, 800*600/8 );
        }/*if*/
    }/*if*/

    if ( mouse_present ) get_mouse_status ();

    if (++refcount < refresh) return;
    refcount=0;

    for (i=0; i<16; i++)
    {
            if ((w = LM_UW(&color[i])) != old_color[i])
            {
                    pal = 1;        /* palette has changed */
                    old_color[i] = w;
            }
    }

    if (pal) change_colors();       /* update the window colormap */

    if (old_shiftmod != shiftmod)
    {
        old_shiftmod = shiftmod;
        change_mode();
    }/*if*/

    p = MEM(vbase);

    if ( !custom_screensize )
    {
        switch ( old_shiftmod )
        {
            case 0:
                outp ( 0x3c4, 2); outp ( 0x3c5, 1 );  // 1. Plane selektieren
                lowres ( p, (char *) 0xa0000, 4000 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 2 );  // 2. Plane selektieren
                lowres ( p+2, (char *) 0xa0000, 4000 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 4 );  // 3. Plane selektieren
                lowres ( p+4, (char *) 0xa0000, 4000 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 8 );  // 4. Plane selektieren
                lowres ( p+6, (char *) 0xa0000, 4000 );
                break;
            case 1:
                outp ( 0x3c4, 2); outp ( 0x3c5, 1 );  // 1. Plane selektieren
                medres ( p, (char *) 0xa0000, 8000 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 2 );  // 2. Plane selektieren
                medres ( p+2, (char *) 0xa0000, 8000 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 1 );
                break;
            case 2:
                hires ( p, (char *) 0xa0000+40*640/8, 640*400/8 );
                break;
        }
    }
    else {
        switch ( old_shiftmod )
        {
            case 0:
                outp ( 0x3c4, 2); outp ( 0x3c5, 1 );  // 1. Plane selektieren
                lowres ( p, (char *) 0xa0000, 800*600 / 16 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 2 );  // 2. Plane selektieren
                lowres ( p+2, (char *) 0xa0000, 800*600 / 16 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 4 );  // 3. Plane selektieren
                lowres ( p+4, (char *) 0xa0000, 800*600 / 16 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 8 );  // 4. Plane selektieren
                lowres ( p+6, (char *) 0xa0000, 800*600 / 16 );
                break;
            case 1:
                outp ( 0x3c4, 2); outp ( 0x3c5, 1 );  // 1. Plane selektieren
                medres ( p, (char *) 0xa0000, 800*600 / 16 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 2 );  // 2. Plane selektieren
                medres ( p+2, (char *) 0xa0000, 800*600 / 16 );
                outp ( 0x3c4, 2); outp ( 0x3c5, 1 );
                break;
            case 2:
                outp ( 0x3c4, 2); outp ( 0x3c5, 1 );  // Plane 1 selektieren
                hires ( (char *) p, (char *) 0xa0000, 800*600/8 );
        }
    }
}

void dos_init_keys ( void ) {
}