/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 *
 * BUGS:
 * - vro_cpyfm probabyl doesn't work properly when fancy modes are used for
 *   screen->memory copying
 * - user-defined fill patterns are global (?)
 *
 * - need to 1) debug and 2) optimize the raster ops (in that order)
 *
 * - should use own vqt_attributes() if custom screen size looks better with
 *   larger font (really? maybe I just need to patch the appropriate variables)
 */
#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
// #include <X11/Xlib.h>
// #include <X11/Xutil.h>
#include "mem.h"
#include "utils.h"
#include "screen.h"
#include "xlib_vdi.h"
#include "debug.h"
#include "main.h"

SCRDEF scr_def[3];

// extern Display *display;

#ifndef TRACE_VDI
#define TRACE_VDI 0
#endif
#define UPDATE_PHYS 0
#define DUAL_MODE 1
#define CALL_ROM_VDI !(UPDATE_PHYS || DUAL_MODE)

#if TRACE_VDI
#define V(_x) debug _x
static UL last_pblock;
#else
#define V(_x) 
#endif


#if XBUFFER
#define XBUF(_x) _x
#else
#define XBUF(_x)
#endif



/* Dual mode: must execute all attribute functions etc. in TOS too!


 */

#define MAX_VWK 100
#define MAX_POINTS 1000
#define MAX_VDI_COLS	256

#define FIX_COLOR(_c) (priv_cmap ? vdi_maptab[(_c)&(MAX_VDI_COLS-1)]\
					: mapcol[vdi_maptab[(_c)&(MAX_VDI_COLS-1)]])

#define LAST_SCREEN_DEVICE 10

#define CHAR_HEIGHT() ((vdi && scr_height >= 400) || scr_planes==1 ? 16 : 8)


#define DEV_LOCATOR 1
#define DEV_VALUATOR 2
#define DEV_CHOICE 3
#define DEV_STRING 4
#define MODE_UNDEFINED 0
#define MODE_REQUEST 1
#define MODE_SAMPLE 2

#define WHITE 0
#define BLACK 1

extern char mapcol[];
unsigned int black_pixel, white_pixel; /* XXX */

UL abase;
UL fontbase;

static int vdi_planes;

#define SYSFONTS 3

#define IGNORE(_x) static int _x(void){V(("%s[%d] ignored\n", #_x, V_HANDLE )); return FALSE; }
#define NYI(_x) static int _x(void){V(("%s[%d] NOT YET IMPLEMENTED\n", #_x, V_HANDLE )); return FALSE; }

#define vdi_w scr_width
#define vdi_h scr_height

/*
    TODO:

Opcode      VDI Name                   Function

    1       v_opnwk                    Open Workstation
    2       v_clswk                    Close Workstation
    3       v_clrwk                    Clear Workstation
    4       v_updwk                    Update Workstation
    5       v_escape                   Escape Functions
        1   vq_chcells                 Inquire addressable character cells
        2   v_exit_cur                 Exit alpha mode
        3   v_enter_cur                Enter alpha mode
        4   v_curup                    Cursor up
        5   v_curdown                  Cursor down
        6   v_curright                 Cursor right
        7   v_curleft                  Cursor left
        8   v_curhome                  Home Cursor
        9   v_eeos                     Erase to end of screen
       10   v_eeol                     Erase to end of line
       11   vs_curaddress              Direct cursor address
       12   v_curtext                  Output cursor addressable text
       13   v_rvon                     Start reverse video
       14   v_rvoff                    End reverse video
       15   vq_curaddress              Inquire current alpha cursor address
       16   vq_tabstatus
       17   v_hardcopy
       18   v_dspcur                   Place graphic cursor
       19   v_rmcur                    Remove last graphic cursor
       20   v_form_adv                 Eject page
       21   v_output_window
       22   v_clear_disp_list
       23   v_bit_image
       61   v_sound
       62   vs_mute
       77   vq_calibrate
       81   vt_resolution
       82   vt_axis
       83   vt_origin
       84   vq_tdimensions
       85   vt_alignment
       98   v_meta_extents
       99   v_write_meta
       99,0 vm_pagesize
       99,1 vm_coords
       100  vm_filename
       101  v_offset
       102  v_fontinit
       2000 v_escape2000
    6       v_pline                    Polyline
    7       v_pmarker                  Polymarker
    8       v_gtext                    Text
    9       v_fillarea                 Filled area
   11                                  GDP
        1   v_bar                      Bar
        2   v_arc                      Arc
        3   v_pieslice                 Pie
        4   v_circle                   Circle
        5   v_ellipse                  Ellipse
        6   v_ellarc                   Elliptical arc
        7   v_ellpie                   Elliptical pie
        8   v_rbox                     Rounded rectangle
        9   v_rfbox                    Filled rounded rectangle
       10   v_justified                Justified graphics text
    12      vst_height                 Set character height, absolute mode
    13      vst_rotation               Set text rotation
    14      vs_color                   Set color representation
    15      vsl_type                   Set polyline linetype
    16      vsl_width                  Set polyline width
    17      vsl_color                  Set polyline color index
    18      vsm_type                   Set polymarker type
    19      vsm_height                 Set polymarker height
    20      vsm_color                  Set polymarker color index
    21      vst_font                   Set text face
    22      vst_color                  Set graphics text color index
    23      vsf_interior               Set fill interior style
    24      vsf_style                  Set fill style index
    25      vsf_color                  Set fill color index
    26      vq_color                   Inquire color representation
    27      vq_cellarray               Inquire color lookup table
    28      vrq_locator,vsm_locator    Input locator
    29      vrq_valuator,vsm_valuator  
    30      vrq_choice,vsm_choice
    31      vrq_string,vsm_string      Input string
    32      vswr_mode                  Set writing mode
    33      vsin_mode                  Set input mode
    34
    35      vql_attribues              Inquire current polyline attributes
    36      vqm_attributes             Inquire current polymarker attributes
    37      vqf_attributes             Inquire current fill area attributes
    38      vqt_attributes             Inquire current graphic text attributes
    39      vst_alignment              Set graphic text alignment
   100      v_opnvwk                   Open virtual screen workstation
   101      v_clsvwk                   Close virtual screen workstation
   102      vq_extnd                   Extended inquire function
   103      v_contourfill
   104      vsf_perimeter              Set fill perimeter visibility
   105      v_get_pixel
   106      vst_effects                Set graphic text special effects
   107      vst_point                  Set character cell height, points mode
   108      vsl_ends                   Set polyline end styles
   109      vro_cpyfm                  Copy raster, opaque
   110      vr_trnfm                   Transform form
   111      vsc_form                   Set mouse form
   112      vsf_udpat                  Set user-defined fill pattern
   113      vsl_udsty                  Set user-defined linestyle
   114      vr_recfl                   Fill rectangle
   115      vqin_mode                  Inquire input mode
   116      vqt_extent                 Inquire text extent
   117      vqt_width                  Inquire character cell width
   118      vex_timv                   Exchange timer interrupt vector
   119      vst_load_fonts
   120      vst_unload_fonts
   121      vrt_cpyfm                  Copy raster, transparent
   122      v_show_c                   Show cursor
   123      v_hide_c                   Hide cursor
   124      vq_mouse                   Sample mouse button state
   125      vex_butv                   Exchange button change vector
   126      vex_motv                   Exchange mouse movement vector
   127      vex_curv                   Exchange cursor change vector
   128      vq_key_s                   Sample keyboard state information
   129      vs_clip                    Set clipping rectangle
   130      vqt_name                   Inquire face name and index
   131      vqt_fontinfo               Inquire current face information

*/

static UL control, intin, ptsin, intout, ptsout;

#define CONTRL(_x) MEM(control+2*(_x))
#define INTIN(_x) MEM(intin+2*(_x))
#define PTSIN(_x) MEM(ptsin+2*(_x))
#define INTOUT(_x) MEM(intout+2*(_x))
#define PTSOUT(_x) MEM(ptsout+2*(_x))

#define V_OPCODE     LM_W(CONTRL(0))
#define V_NPTSIN     LM_W(CONTRL(1))
#define V_NPTSOUT(x) SM_W(CONTRL(2),x)
#define V_NINTIN     LM_W(CONTRL(3))
#define V_NINTOUT(x) SM_W(CONTRL(4),x)
#define V_SUBCODE    LM_W(CONTRL(5))
#define V_HANDLE     LM_W(CONTRL(6))

#define MD_REPLACE 1
#define MD_TRANS   2
#define MD_XOR     3
#define MD_ERASE   4

#define FIS_HOLLOW  0
#define FIS_SOLID   1
#define FIS_PATTERN 2
#define FIS_HATCH   3
#define FIS_USER    4

NYI(nothing)

void vdi_redraw(int x, int y, int w, int h)
{
}


static void catchsig(int sig)
{
}

void vdi_init(void)
{
}


void vdi_exit(void)
{
}


static void destroy_image(void)
{
}


static void change_mode(void)
{
}


static void clear_window(void)
{
}


static void init_window(void)
{
}


int Vdi(void)
{
	return FALSE;
}


static void fix_vars(int width, int height, int planes)
{
	UW *a;

	T(("Fixing Line-A variables... %dx%dx%d\n", width, height, planes));
	a = (UW *)MEM(abase);
	SM_W(a + 1, (width * planes) / 8);
	SM_W(a - 1, (width * planes) / 8);
	SM_W(a - 23, CHAR_HEIGHT());
	SM_W(a - 22, width / 8 - 1);
	SM_W(a - 21, height / CHAR_HEIGHT() - 1);
	SM_W(a - 20, ((width * planes) / 8) * CHAR_HEIGHT());
	SM_W(a - 6, width);
	SM_W(a - 2, height);
	SM_W(a - 0x2b4 / 2, width - 1);
	SM_W(a - 0x2b2 / 2, height - 1);
#if 0
	fprintf(stderr, "FBASE=%08x, CUR_FONT=%08x, DEF_FONT=%08x, [%d,%d,%d,%d]\n",
		LM_UL(a + 0x54 / 2), LM_UL(a - 0x38a / 2), LM_UL(a - 0x1cc / 2),
		LM_W(a - 0x1f2 / 2), LM_W(a - 0x1f0 / 2), LM_W(a - 0x1ee / 2), LM_W(a - 0x1ec / 2));
	if (CHAR_HEIGHT() == 16)
	{
		SM_UL(a - 0x1cc / 2, LM_UL(MEM(fontbase + 8)));
		SM_UL(a - 0x38a / 2, LM_UL(MEM(fontbase + 8)));
	}
	fprintf(stderr, "-> FBASE=%08x, CUR_FONT=%08x, DEF_FONT=%08x\n",
		LM_UL(a + 0x54 / 2), LM_UL(a - 0x38a / 2), LM_UL(a - 0x1cc / 2));
#endif
}

void vdi_printtab(void)
{
	int i;
	
	V(("vdi_maptab: (%d/%d planes)\n", scr_planes, vdi_planes));
	for (i = 0; i < (1 << scr_planes); i++)
	{
		V(("%3d -> %3d  %8lx\n", i, vdi_maptab[i], mapcol[vdi_maptab[i]] ));
	}
}

void vdi_post(void)
{
	UW *a;
	UL pblock;
	
	SP += 6;
	pblock = LM_UL(MEM(SP));
	pc = pblock;
	SP += 4;
	pblock = LM_UL(MEM(SP));
	SP += 4;
	
	control = LM_UL(MEM(pblock));
	if (V_OPCODE != 1 && V_OPCODE != 100)
	{
		return;
	}
	intin = LM_UL(MEM(pblock + 4));
	ptsin = LM_UL(MEM(pblock + 8));
	intout = LM_UL(MEM(pblock + 12));
	ptsout = LM_UL(MEM(pblock + 16));
	a = (UW *)MEM(abase);
	vdi_planes = LM_W(a);
	if (fix_screen)
	{
		SM_W(INTOUT(0), vdi_w - 1);
		SM_W(INTOUT(1), vdi_h - 1);
		fix_vars(vdi_w, vdi_h, vdi_planes);
	}
	if (!vdi)
		return;
	
}

void Init_Linea(void)
{
	abase = DREG(0);
	fontbase = AREG(1);
	fprintf(stderr,"Initializing Line-A... (abase=%08lx)\n", (long)abase);
	if (fix_screen)
		fix_vars(scr_width, scr_height, scr_planes);
}
